/*
 *  A simple X11 driver
 *  Copyright (C) 2002 BlueEyedOS
 *  This program is freeware (and bugware)
 *  Hardcoded with a display of 1024*768 32bits
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <support/SupportDefs.h>
#include <cosmoe/isa_io.h>

#include <cosmoe/vesa_gfx.h>
#include <cosmoe/kernel.h>
#include "sfont.h"
#include <interface/Bitmap.h>

#include "GMdrv.h"
#include "bitmap.h"
#include "inputnode.h" 
#include "server.h" 
#include "keyboard.h"

using namespace os;

inline int MIN(int a,int b)
{
  if(a<b)
    return a;
  return b;
}
inline int MAX(int a,int b)
{
  if(b<a)
    return a;
  return b;
}

static area_id        g_nFrameBufArea = -1;

GMDriver::GMDriver() :m_nCurrentMode(0),
		      m_nFrameBufferSize(0),
		      m_nFrameBufferOffset(0){
  busy=false;c=0;//bug tracking
  mysem= create_sem_etc( 1, "prevent X crash",0 );
  refreshsem= create_sem_etc( 1, "still crashed!",0 );
}

GMDriver::~GMDriver()
{
}


bool GMDriver::InitModes( void ){
  m_cModeList.push_back( ScreenMode( 1024, 768, 1024*4,CS_RGB32 ) );
  return( true );
}


void event_listener(void *arg)
{
  
  GMDriver *d= (GMDriver *)arg;
  bool stop=false;
  while (!stop)
    { 
      XEvent x_event;
      XNextEvent(d->display, &x_event);
      acquire_sem(d->mysem);
      switch  (x_event.type) 
	{
	case MotionNotify:{
	  int x=((XMotionEvent *) &x_event)->x;
	  int y=((XMotionEvent *) &x_event)->y;
	  //printf("MouseMoved (%d,%d)\n",x,y);
	  InputNode::SetMousePos( CPoint(x,y) );
	  break;
	}
	case Expose:
	  //printf ("Draw\n");
	  d->Refresh(0,0,1024,768);
	  break; 
	case KeyPress:{
	  //printf ("KeyPDown\n");
	  char k=XLookupKeysym(&x_event.xkey, 0);
	  char* key=(char*)malloc(2*sizeof(char));
	  key[0]=k;
	  key[1]=0;
	  // hack :)
	  SrvWindow::SendKbdEvent(((XKeyEvent *) &x_event)->keycode,0,key,key);
	  break;
	}
	case KeyRelease:{
	  //printf ("KeyUp\n");
	  break;
	}

	case ButtonPress:{
	  //printf ("MouseDown\n");
	  CMessage*  pcEvent = new CMessage( M_MOUSE_DOWN );
	  pcEvent->AddInt32( "buttons",  ((XButtonEvent *) &x_event)->button);
	  InputNode::EnqueueEvent( pcEvent );
	  break;
	}
	case ButtonRelease:{
	  //printf ("MouseUp\n");
	  CMessage*  pcEvent = new CMessage( M_MOUSE_UP );
	  pcEvent->AddInt32( "buttons", ((XButtonEvent *) &x_event)->button );
	  InputNode::EnqueueEvent( pcEvent );
	  break;
	}
	default:	  
	  break;
	  
	}
      XFlush(d->display);
      release_sem(d->mysem);
    }
}


area_id GMDriver::Open( void )
{
  printf("GMDriver::Open\n");
  display = XOpenDisplay (NULL);
  printf("GMDriver::display:%d\n",(int)display);
  screen = DefaultScreen (display);
  depth = DefaultDepth (display, screen);
  window_attributes.border_pixel = BlackPixel (display, screen);
  window_attributes.background_pixel = WhitePixel (display, screen);
  window_attributes.override_redirect = True;
  window_mask = CWBackPixel | CWBorderPixel;
  
  xcanvas = XCreateWindow(display, DefaultRootWindow(display), 
			  0, 0, 1024, 768, 0, depth, 
			  InputOutput, CopyFromParent, 
			  CWBackPixel | CWBorderPixel /*if you need noborder */
			  /*| CWOverrideRedirect*/, &window_attributes);
  image_gc = XCreateGC (display, xcanvas, 0, 0);
  XMapWindow (display, xcanvas);
  XFlush(display);
  XSelectInput(display, xcanvas, ExposureMask | PointerMotionMask| 
	       KeyPressMask | ButtonPressMask  | ButtonReleaseMask | StructureNotifyMask);
  printf("XCreateWindow passed\n");
 
/* 
    bool stop=false;
   // nice to wait XMapEvent if no CWOverrideRedirect
  while (!stop)
    {
      XEvent x_event;
      XNextEvent(display, &x_event);
      printf("X\n");
      switch  (x_event.type) 
	{
	case MapNotify:
	  stop=true;
	  break;
	default:
	  printf("waiting the XWindow\n");
	  break;
	  
	}
    }
 */


  if ( InitModes() )
	{
	  printf("GMDriver::Open initDone \n");
	  m_nFrameBufferSize = 1024 * 768 * 4*4;
	  /*      m_pFrameBuffer = NULL;*/
	  buffer_32bpp=(char*)malloc(m_nFrameBufferSize);
	 
	  g_nFrameBufArea = create_area( "X11_fb",(void**) &buffer_32bpp, m_nFrameBufferSize,AREA_FULL_ACCESS, AREA_NO_LOCK );
	  
	  printf("GMDriver::Open initDone area:%d\n",	g_nFrameBufArea);
	  xim = XCreateImage (display, CopyFromParent, depth, ZPixmap, 0,
			      (char *) buffer_32bpp, 1024,
			      768, 32, 1024 *4);
	  
	  
	  mon_pixmap = XCreatePixmap(display, xcanvas, 1024, 768, depth);               
	  XPutImage (display, mon_pixmap, image_gc, xim, 0, 0, 0, 0, 1024,768);
	  XFlush(display);
	  
	  /// New thread for Mouse Events
	  pthread_t p_thread;
	  pthread_create (&p_thread, NULL, (void *(*) (void *))&event_listener,(void *) this);


	  return( g_nFrameBufArea );
	}
	return( -1 );
}


void GMDriver::Close( void )
{
}

int GMDriver::GetScreenModeCount()
{
	return( m_cModeList.size() );
}


bool GMDriver::GetScreenModeDesc( int nIndex, ScreenMode* psMode )
{
	if ( nIndex >= 0 && nIndex < int(m_cModeList.size()) )
	{
		*psMode = m_cModeList[nIndex];
		return( true );
	}

	return( false );
}


int GMDriver::SetScreenMode( int nWidth, int nHeight, ColorSpace eColorSpc,
			     int nPosH, int nPosV, int nSizeH, int nSizeV, float vRefreshRate )
{
  return 0; 
}

int GMDriver::GetHorizontalRes()
{
  return( m_cModeList[m_nCurrentMode].m_nWidth );
}

int GMDriver::GetVerticalRes()
{
  return( m_cModeList[m_nCurrentMode].m_nHeight );
}

int GMDriver::GetBytesPerLine()
{
  return( m_cModeList[m_nCurrentMode].m_nBytesPerLine );
}

color_space GMDriver::GetColorSpace()
{
  return( m_cModeList[m_nCurrentMode].m_eColorSpace );
}


bool GMDriver::IntersectWithMouse( const IRect& cRect )
{
  return( false );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void GMDriver::SetColor( int nIndex, const Color32_s& sColor )
{
  // lol
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool GMDriver::SetVesaMode( uint32 nMode )
{
  return ( set_vesa_mode( nMode ) == 0 );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool GMDriver::DrawLine( SrvBitmap* psBitMap, const IRect& cClipRect,
			 const IPoint& cPnt1, const IPoint& cPnt2, const Color32_s& sColor, int nMode )
{
  acquire_sem(mysem);
  
  DisplayDriver::DrawLine( psBitMap, cClipRect, cPnt1, cPnt2, sColor, nMode );
  int minX=MIN(cPnt1.x, cPnt2.x);
  int minY=MIN(cPnt1.y, cPnt2.y);
  int maxX=MAX(cPnt1.x, cPnt2.x);
  int maxY=MAX(cPnt1.y, cPnt2.y);

  Refresh( minX , minY, 
	    MAX(0,maxX-minX+1),
	   MAX(0,maxY-minY+1));
 
  release_sem(mysem);
 return( true );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool GMDriver::FillRect( SrvBitmap* psBitMap, const IRect& cRect, const Color32_s& sColor )
{
  acquire_sem(mysem);

  DisplayDriver::FillRect( psBitMap, cRect, sColor );
  Refresh(cRect.left , cRect.top, 
	  cRect.right-cRect.left+1,cRect.bottom-cRect.top+1);
 
  release_sem(mysem);
  return true;   
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool GMDriver::BltBitmap( SrvBitmap* dstbm, SrvBitmap* srcbm, IRect cSrcRect, IPoint cDstPos, int nMode )
{
  acquire_sem(mysem);
  

  int width,height;
 
  DisplayDriver::BltBitmap( dstbm, srcbm, cSrcRect, cDstPos, nMode );
	
  width= cSrcRect.right-cSrcRect.left+1;
  height=cSrcRect.bottom-cSrcRect.top+1;
  
  Refresh(cDstPos.x,cDstPos.y,width,height);
  
  release_sem(mysem);
  return(true  ); 
}

bool GMDriver::RenderGlyph( SrvBitmap *pcBitmap, Glyph* pcGlyph,
			    const IPoint& cPos, const IRect& cClipRect, const Color32_s& sFgColor )
{ 
  acquire_sem(mysem);

 
  IRect        cBounds        = pcGlyph->m_cBounds + cPos;
  IRect        cRect         = cBounds & cClipRect;
  
  if ( cRect.IsValid() == false )
    {
      c--;
      release_sem(mysem);
      return true;
    }
  
  
  int sx = cRect.left - cBounds.left;
  int sy = cRect.top - cBounds.top;
  
  int nWidth        = cRect.Width()+1;
  int nHeight        = cRect.Height()+1;
  
  int        nSrcModulo = pcGlyph->m_nBytesPerLine - nWidth;
  
  uint8*  pSrc = pcGlyph->m_pRaster + sx + sy * pcGlyph->m_nBytesPerLine;
  
  Color32_s        sCurCol;
  Color32_s        sBgColor;
  
  if ( pcBitmap->m_eColorSpc == CS_RGB15 )
    {
      int        nDstModulo = pcBitmap->m_nBytesPerLine / 2 - nWidth;
      uint16* pDst = (uint16*)pcBitmap->m_pRaster + cRect.left + (cRect.top * pcBitmap->m_nBytesPerLine / 2);
      
      int nFgClut = COL_TO_RGB15( sFgColor );
      
      for ( int y = 0 ; y < nHeight ; ++y )
	{
	  for ( int x = 0 ; x < nWidth ; ++x )
	    {
	      int nAlpha = *pSrc++;
	      
	      if ( nAlpha > 0 ) {
		if ( nAlpha == NUM_FONT_GRAYS - 1 )
		  {
		    *pDst = nFgClut;
		  } else {
		    int        nClut = *pDst;
		    
		    sBgColor = RGB16_TO_COL( nClut );
		    
		    
		    sCurCol.red   = sBgColor.red   + (sFgColor.red   - sBgColor.red)   * nAlpha / (NUM_FONT_GRAYS-1);
		    sCurCol.green = sBgColor.green + (sFgColor.green - sBgColor.green) * nAlpha / (NUM_FONT_GRAYS-1);
		    sCurCol.blue  = sBgColor.blue  + (sFgColor.blue  - sBgColor.blue)  * nAlpha / (NUM_FONT_GRAYS-1);
		    
		    *pDst = COL_TO_RGB15( sCurCol );
		  }
	      }
	      pDst++;
	    }
	  pSrc += nSrcModulo;
	  pDst += nDstModulo;
	}
    }
  else if ( pcBitmap->m_eColorSpc == CS_RGB16 )
    {
      int        nDstModulo = pcBitmap->m_nBytesPerLine / 2 - nWidth;
      uint16* pDst = (uint16*)pcBitmap->m_pRaster + cRect.left + cRect.top * pcBitmap->m_nBytesPerLine / 2;
      
      int nFgClut = COL_TO_RGB16( sFgColor );
      
      for ( int y = 0 ; y < nHeight ; ++y )
	{
	  for ( int x = 0 ; x < nWidth ; ++x )
	    {
	      int nAlpha = *pSrc++;
	      
	      if ( nAlpha > 0 ) {
		if ( nAlpha == NUM_FONT_GRAYS - 1 )
		  {
		    *pDst = nFgClut;
		  } else {
		    int        nClut = *pDst;
		    
		    sBgColor = RGB16_TO_COL( nClut );
		    
		    sCurCol.red   = sBgColor.red   + int(sFgColor.red   - sBgColor.red)   * nAlpha / (NUM_FONT_GRAYS-1);
		    sCurCol.green = sBgColor.green + int(sFgColor.green - sBgColor.green) * nAlpha / (NUM_FONT_GRAYS-1);
		    sCurCol.blue  = sBgColor.blue  + int(sFgColor.blue  - sBgColor.blue)  * nAlpha / (NUM_FONT_GRAYS-1);
		    
		    *pDst = COL_TO_RGB16( sCurCol );
		  }
	      }
	      pDst++;
	    }
	  pSrc += nSrcModulo;
	  pDst += nDstModulo;
	}
    }
  else if ( pcBitmap->m_eColorSpc == CS_RGB32 )
    {
      int        nDstModulo = pcBitmap->m_nBytesPerLine / 4 - nWidth;
      uint32* pDst = (uint32*)pcBitmap->m_pRaster + cRect.left + cRect.top * pcBitmap->m_nBytesPerLine / 4;
      
      int nFgClut = COL_TO_RGB32( sFgColor );
      
      for ( int y = 0 ; y < nHeight ; ++y )
	{
	  for ( int x = 0 ; x < nWidth ; ++x )
	    {
	      int nAlpha = *pSrc++;
	      
	      if ( nAlpha > 0 ) {
		if ( nAlpha == NUM_FONT_GRAYS - 1 )
		  {
		    *pDst = nFgClut;
		  } else {
		    int        nClut = *pDst;
		    
		    sBgColor = RGB32_TO_COL( nClut );

		    sCurCol.red   = sBgColor.red   + (sFgColor.red   - sBgColor.red)   * nAlpha / (NUM_FONT_GRAYS-1);
		    sCurCol.green = sBgColor.green + (sFgColor.green - sBgColor.green) * nAlpha / (NUM_FONT_GRAYS-1);
		    sCurCol.blue  = sBgColor.blue  + (sFgColor.blue  - sBgColor.blue)  * nAlpha / (NUM_FONT_GRAYS-1);
		    
		    *pDst = COL_TO_RGB32( sCurCol );
		  }
	      }
	      pDst++;
	    }
	  pSrc += nSrcModulo;
	  pDst += nDstModulo;
	}
    }
  
  Refresh(cRect.left , cRect.top,cRect.right-cRect.left+1,cRect.bottom-cRect.top+1);
  
  release_sem(mysem);
  return(true  );
}
bool GMDriver::RenderGlyph( SrvBitmap *pcBitmap, Glyph* pcGlyph,
                                 const IPoint& cPos, const IRect& cClipRect, uint32* anPalette )
{ 
  RenderGlyph(pcBitmap, pcGlyph,cPos, cClipRect, Color32_s(0,0,0,0));
  
  return(true);	
}

void GMDriver::Refresh(int posx,int posy, int width,int height){

   acquire_sem(refreshsem);
  
   if(posx<0 || posy<0 || posx+width>1024 || posy+height>768)
     {
       printf("************************************ Error refresh****************************************************\n");
       posx=0;
       posy=0;
       width=MIN(1024,width);
       height=MIN(768,height);
     }
      
   XPutImage (display, xcanvas, image_gc, xim, 
	      posx, posy, 
	      posx , posy, 
	      width,height);
  
   release_sem(refreshsem);
}

